unit FIN;

interface

uses unit1;


type
    TTempBuffer = array[1..32768] of byte;
    PTempBuffer = ^TTempBuffer;

var
   InBuffer, OutBuffer: PTempBuffer;
   bytesleft: Cardinal;

procedure CompressFin;
procedure DecompressFin;

implementation

//Buufering rotines to make this algorithim faster
//sizeof buffer = 32768 (32k)
procedure ReadBuffer(var Buf; var bytes_got: integer);
Const
     BytesInMem: integer = 0;
     InBufferPos: integer = 1;
var
   BytesToRead: Cardinal;
begin
     if BytesInMem = 0 then
        begin
             if BytesLeft >= DicSize then
                BytesToRead := DicSize
             else
                 BytesToRead := BytesLeft;

             BlockRead(infile,InBuffer^[1],BytesToRead,BytesInMem);
             Dec(BytesLeft,BytesInMem);

             if BytesInMem = 0 then
                begin
                     bytes_got := 0;
                     Exit;
                end;
             InBufferPos := 1;
        end;
     move(inbuffer^[InBufferPos],Buf,1);
     bytes_got := 1;
     Dec(BytesInMem);
     inc(InBufferPos,1);
     if compressing then
        UpdateCrc(CalCrc,Byte(Buf));
end;

procedure WriteBuffer(ForceFlush: Boolean; buf: byte);
const
     OutBufferPos: integer = 1;
begin
     if ForceFlush then //if we what to force the fluch of the buffer(when compression ends)
        begin
             BlockWrite(Outfile,Outbuffer^[1],OutbufferPos-1);
             //make sure that the next time we use this we are at the begin of the buffer
             OutbufferPos := 1;
             Exit;
        end;
     if OutbufferPos > 32768 then
        begin
             BlockWrite(Outfile,Outbuffer^[1],32768);
             OutbufferPos := 1;
        end;
     Move(buf,Outbuffer^[OutbufferPos],1);
     Inc(OutbufferPos);
     if Compressing then
        //inc the compressed size by 1
        Inc(LocalHeader.CompressSize);
     if not Compressing then
        //calc the Crc
        UpdateCrc(CalCrc,Buf);
end;


{****************************** Encode/Decode FIN ******************************}
type
    tpcTable = array[0..32768] of byte;
    ppcTable = ^TpcTable;

function INDEX(var p1,p2): integer;
begin
   Result := ( Byte(p1) shl 7) xor Byte(p2);
end;

procedure CompressFin;
var
   c: byte;                  // character
   i: integer;               // loop counter
   p1, p2: byte;             // previous 2 characters
   buf: array[0..8] of byte; // keeps characters temporarily
   ctr: integer;             // number of characters in mask
   bctr: integer;            // position in buf
   mask: byte;               // mask to mark successful predictions

   bytesread: integer;
   pcTable: ppcTable;
begin
     p1 := 0;
     p2 := 0;
     ctr := 0;
     bctr := 0;
     mask := 0;
     new(pcTable);
     New(InBuffer);
     New(OutBuffer);
     FillChar(pcTable^,32768,' ');

     ReadBuffer(c,bytesread);
     while bytesread <> 0 do
     begin
          // try to predict the next character
          if pcTable^[INDEX(p1,p2)] = c then
             // correct prediction, mark bit for correct prediction
             mask := mask xor (1 shl ctr)
          else
              begin
                   // wrong prediction, but next time ...
                   pcTable^[INDEX(p1,p2)] := c;
                   // buf keeps character temporarily in buffer
                   buf[bctr] := c;
                   inc(bctr);
              end;
          // test if mask is full (8 characters read)
          inc(ctr);
          if ctr= 8 then
             begin
                  // write mask
                  WriteBuffer(false,mask);
                  // write kept characters
                  for i:=0 to bctr-1 do
                      WriteBuffer(false,buf[i]);
                  // reset variables
                  ctr :=0;
                  bctr:=0;
                  mask:=0;
             end;
          // shift characters
          p1 := p2;
          p2 := c;
          ReadBuffer(c,bytesread);
     end;
     // EOF, but there might be some left for output
     if ctr >= 1 then
        begin
           // write mask
           WriteBuffer(false,mask);
           // write kept characters
           for i:=0 to bctr-1 do
               WriteBuffer(false,buf[i]);
        end;
     dispose(pcTable);
     //clean the buffer
     WriteBuffer(true,1);
     dispose(InBuffer);
     dispose(OutBuffer);
end;

procedure DecompressFin;
var
   ci,co: byte;     // characters (in and out)
   p1,p2: shortint; // previous 2 characters
   ctr: integer;    // number of characters processed for this mask
   mask: byte;      // mask to mark successful predictions

   bytesread: integer;
   pcTable: ppcTable;
begin
     p1 := 0;
     p2 := 0;
     ctr := 8;
     mask := 0;
     New(InBuffer);
     New(OutBuffer);
     New(pcTable);
     fillchar(pcTable^, 32768, ' '); // space (ASCII 32) is the most used char
     ReadBuffer(ci,bytesread);
     while bytesread <> 0 do
     begin
          // get mask (for 8 characters)
          mask := byte(shortint(ci));
          // for each bit in the mask
          for ctr:=0 to 7 do
              begin
                   if (mask and (1 shl ctr)) >= 1 then
                      // predicted character
                      co := pcTable^[INDEX(p1,p2)]
                   else
                       begin
                            // not predicted character
                            ReadBuffer(co,bytesread);
                            if bytesread = 0 then
                            begin
                                 WriteBuffer(true,1);// decompression completed! Clean the buffer and exit
                                 Exit;
                            end;
	                    pcTable^[INDEX(p1,p2)] := byte(co);
                       end;
                   WriteBuffer(false,co);
                   p1 := p2;
                   p2 := co;
              end;
          ReadBuffer(ci,bytesread);
   end;
   Dispose(pcTable);
   //clear the buffer
   WriteBuffer(true,1);

   dispose(InBuffer);
   dispose(OutBuffer);
end;

end.
